# ##################################################################################### #
# ISIS File Polling Repository : https://github.com/ISISSoftwareServices/ISISFilePolling
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# ##################################################################################### #
"""
End of run monitor. Detects new runs that arrive and
sends them off to the autoreduction service.
"""

import copy
import csv
import logging
from typing import Optional

from filelock import FileLock, Timeout
import requests
from requests.models import Response

from autoreduce_run_detection.settings import LOCAL_CACHE_LOCATION, AUTOREDUCE_API_URL, AUTOREDUCE_TOKEN, TEAMS_URL

# pylint:disable=abstract-class-instantiated

LOGGING = logging.getLogger(__package__)

TEAMS_CARD_DATA = {
    "@context": "https://schema.org/extensions",
    "@type": "MessageCard",
    "themeColor": "0072C6",
    "title": "Alert Raised",
    "text": "",
    "potentialAction": []
}


class InstrumentMonitorError(Exception):
    """
    Any fatal exception that occurs during execution of the
    instrument monitor
    """


class InstrumentMonitor:
    """
    Checks the ISIS archive for new runs on an instrument and submits them to ActiveMQ
    """
    def __init__(self,
                 instrument_name: str,
                 last_run_file: str = "",
                 summary_file: str = "",
                 data_dir: str = "",
                 file_ext: str = "",
                 teams_url: Optional[str] = None):
        self.instrument_name = instrument_name
        self.last_run_file = last_run_file
        self.summary_file = summary_file
        self.data_dir = data_dir
        self.file_ext = file_ext
        self.teams_url = teams_url

    def read_instrument_last_run(self):
        """
        Read the last run recorded by the instrument from its lastrun.txt

        Returns:
            Last run on the instrument as a string
        """
        with open(self.last_run_file, 'r') as last_run:
            line_parts = last_run.readline().split()
            if len(line_parts) != 3:
                raise InstrumentMonitorError("Unexpected last run file format for '{}'".format(self.last_run_file))
        return line_parts

    def submit_runs(self, start_run, end_run) -> Response:
        """
        Submit a run via the REST API

        Args:
            summary_rb_number: RB number of the experiment as read from the summary file
            run_number: Run number as it appears in lastrun.txt
            file_name: File name e.g. GEM1234.nxs
        """
        # Check to see if the last run exists, if not then raise an exception
        LOGGING.info(
            "Submitting runs in range %i - %i for %s",
            start_run,
            end_run,
            self.instrument_name,
        )
        try:
            response = requests.post(
                AUTOREDUCE_API_URL.format(instrument=self.instrument_name),
                json={
                    "runs": list(range(start_run, end_run + 1)) if abs(end_run - start_run) > 0 else [start_run],
                    "user_id": 0  # AUTOREDUCTTION_SERVICE user id
                },
                headers={
                    "Content-Type": "application/json",
                    "Authorization": f"Token {AUTOREDUCE_TOKEN}"
                })

            if response.status_code != 200:
                LOGGING.error("Request error when submitting runs in range %i - %i for %s, error: %s", start_run,
                              end_run, self.instrument_name, response.content)
                raise InstrumentMonitorError(f"Request status code is not 200, error: {response.content}")
            return response
        except requests.exceptions.RequestException as err:
            LOGGING.error("Failed to submit runs %i - %i for instrument %s", start_run, end_run, self.instrument_name)
            if self.teams_url:
                data = copy.deepcopy(TEAMS_CARD_DATA)
                data["text"] = f"Failed to submit runs {start_run} - {end_run} for instrument {self.instrument_name}"
                try:
                    requests.post(self.teams_url, json=data)
                except requests.exceptions.RequestException:
                    LOGGING.error("Failed to send message using this TEAMS url: %s", self.teams_url)
            else:
                LOGGING.info("No TEAMS_URL set, not sending message to Teams")
            raise InstrumentMonitorError() from err

    def submit_run_difference(self, local_last_run):
        """
        Submit the difference between the last run on the archive for this
        instrument
        Args:
            local_last_run: Local last run to check against
        """
        # Get archive lastrun.txt
        last_run_data = self.read_instrument_last_run()
        instrument_last_run = last_run_data[1]

        local_run_int = int(local_last_run)
        instrument_run_int = int(instrument_last_run)

        if instrument_run_int > local_run_int:
            LOGGING.info(self.submit_runs(local_run_int + 1, instrument_run_int + 1))
        return str(instrument_run_int)


def update_last_runs(csv_name):
    """
    Read the last runs CSV file and bring it up to date with the
    instrument lastrun.txt

    Args:
        csv_name: File name of the local last runs CSV file
    """
    # Loop over instruments
    output = []
    with open(csv_name, 'r') as csv_file:
        csv_reader = csv.reader(csv_file)
        for row in csv_reader:
            LOGGING.info("Processing instrument %s with last run %i", row[0], int(row[1]))
            inst_mon = InstrumentMonitor(instrument_name=row[0],
                                         last_run_file=row[2],
                                         summary_file=row[3],
                                         data_dir=row[4],
                                         file_ext=row[5],
                                         teams_url=TEAMS_URL)

            try:
                last_run = inst_mon.submit_run_difference(row[1])
                row[1] = last_run
            except InstrumentMonitorError as ex:
                LOGGING.error(ex)
            output.append(row)

    # Write any changes to the CSV
    with open(csv_name, 'w', newline='') as csv_file:
        csv_writer = csv.writer(csv_file)
        for row in output:
            csv_writer.writerow(row)


def main():
    """
    Ingestion Entry point
    """
    # Acquire a lock on the last runs CSV file to prevent access
    # by other instances of this script
    try:
        with FileLock("{}.lock".format(LOCAL_CACHE_LOCATION), timeout=1):
            update_last_runs(LOCAL_CACHE_LOCATION)
    except Timeout:
        LOGGING.error("Error acquiring lock on last runs CSV." " There may be another instance running.")


if __name__ == '__main__':
    main()  # pragma: no cover



# ##################################################################################### #
# ISIS File Polling Repository : https://github.com/ISISSoftwareServices/ISISFilePolling
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# ##################################################################################### #
"""
Unit tests for run_detection
"""
import csv
import os
from unittest.mock import Mock, patch, call
from unittest import TestCase
from requests.exceptions import RequestException, ConnectionError  # pylint:disable=redefined-builtin

from filelock import FileLock
from parameterized import parameterized

from autoreduce_run_detection.run_detection import InstrumentMonitor, InstrumentMonitorError, update_last_runs, main
from autoreduce_run_detection.settings import AUTOREDUCE_API_URL, LOCAL_CACHE_LOCATION

# pylint:disable=abstract-class-instantiated

# Test data
SUMMARY_FILE = ("WIS44731Smith,Smith,"
                "SmithCeAuSb2 MRSX ROT=15.05 s28-MAR-2019 09:14:23    34.3 1820461\n"
                "WIS44732Smith,Smith,"
                "SmithCeAuSb2 MRSX ROT=15.05 s28-MAR-2019 10:23:47    40.0 1820461\n"
                "WIS44733Smith,Smith,"
                "SmithCeAuSb2 MRSX ROT=15.05 s28-MAR-2019 11:34:25     9.0 1820461\n")
LAST_RUN_FILE = "WISH 00044733 0 \n"
INVALID_LAST_RUN_FILE = "INVALID LAST RUN FILE"
RUN_DICT = {
    'instrument': 'WISH',
    'run_number': '00044733',
    'data': '/my/data/dir/cycle_18_4/WISH00044733.nxs',
    'rb_number': '1820461',
    'facility': 'ISIS',
    'started_by': 0
}
RUN_DICT_SUMMARY = {
    'instrument': 'WISH',
    'run_number': '00044733',
    'data': '/my/data/dir/cycle_18_4/WISH00044733.nxs',
    'rb_number': '1820333',
    'facility': 'ISIS'
}
CSV_FILE = "WISH,44733,lastrun_wish.txt,summary_wish.txt,data_dir,.nxs"
LASTRUN_WISH_TXT = "WISH 44734 0"


# pylint:disable=too-few-public-methods,missing-function-docstring
class DataHolder:
    """
    Small helper class to represent expected nexus data format
    """
    def __init__(self, data):
        self.data = data

    def get(self, _):
        mock_value = Mock()
        mock_value.value = self.data
        return mock_value


# nexusformat mock objects
NXLOAD_MOCK = Mock()
NXLOAD_MOCK.items = Mock(return_value=[('raw_data_1', DataHolder([b'1910232']))])

NXLOAD_MOCK_EMPTY = Mock()
NXLOAD_MOCK_EMPTY.items = Mock(return_value=[('raw_data_1', DataHolder(['']))])


class TestRunDetection(TestCase):
    def tearDown(self):
        if os.path.isfile('test_lastrun.txt'):
            os.remove('test_lastrun.txt')
        if os.path.isfile('test_summary.txt'):
            os.remove('test_summary.txt')
        if os.path.isfile('test_last_runs.csv'):
            os.remove('test_last_runs.csv')
        if os.path.isfile('lastrun_wish.txt'):
            os.remove('lastrun_wish.txt')

    def test_read_instrument_last_run(self):
        with open('test_lastrun.txt', 'w') as last_run:
            last_run.write(LAST_RUN_FILE)

        inst_mon = InstrumentMonitor('WISH')
        inst_mon.last_run_file = 'test_lastrun.txt'
        last_run_data = inst_mon.read_instrument_last_run()

        self.assertEqual('WISH', last_run_data[0])
        self.assertEqual('00044733', last_run_data[1])
        self.assertEqual('0', last_run_data[2])

    # pylint:disable=invalid-name
    def test_read_instrument_last_run_invalid_length(self):
        with open('test_lastrun.txt', 'w') as last_run:
            last_run.write(INVALID_LAST_RUN_FILE)

        inst_mon = InstrumentMonitor('WISH')
        inst_mon.last_run_file = 'test_lastrun.txt'
        with self.assertRaises(InstrumentMonitorError):
            inst_mon.read_instrument_last_run()

    def test_submit_run_difference(self):
        # Setup test
        inst_mon = InstrumentMonitor('WISH')
        inst_mon.submit_runs = Mock(return_value=None)
        inst_mon.file_ext = '.nxs'
        inst_mon.read_instrument_last_run = Mock(return_value=['WISH', '00044733', '0'])

        # Perform test
        run_number = inst_mon.submit_run_difference(44731)
        self.assertEqual(run_number, '44733')
        inst_mon.submit_runs.assert_has_calls([call(44732, 44734)])

    @patch('autoreduce_run_detection.run_detection.requests.post', return_value=[44734, 44735])
    def test_update_last_runs(self, requests_post_mock: Mock):
        # write out the local lastruns.csv that is used to track each instrument
        with open('test_last_runs.csv', 'w') as last_runs:
            last_runs.write(CSV_FILE)

        # write out the lastruns.txt file that would usually be on the archive
        with open('lastrun_wish.txt', 'w') as lastrun_wish:
            lastrun_wish.write(LASTRUN_WISH_TXT)

        # Perform test
        update_last_runs('test_last_runs.csv')
        requests_post_mock.assert_called_once()
        assert requests_post_mock.call_args[0][0] == AUTOREDUCE_API_URL.format(instrument="WISH")
        assert "json" in requests_post_mock.call_args[1]
        assert "headers" in requests_post_mock.call_args[1]

        assert requests_post_mock.call_args[1]["json"]["runs"] == [44734, 44735]
        assert requests_post_mock.call_args[1]["json"]["user_id"] == 0

        # Read the CSV and ensure it has been updated
        with open('test_last_runs.csv') as csv_file:
            csv_reader = csv.reader(csv_file)
            for row in csv_reader:
                if row:  # Avoid the empty rows
                    self.assertEqual('44734', row[1])

    @parameterized.expand([
        [ConnectionError],
        [RequestException],
    ])
    @patch('autoreduce_run_detection.run_detection.requests.post', return_value=[44734, 44735])
    @patch('autoreduce_run_detection.run_detection.LOGGING')
    def test_update_last_runs_with_error(self, exception_class, logger_mock: Mock, requests_post_mock: Mock):
        """
        Test trying to update last runs but the request to the autoreduce API fails.
        """
        # Setup test
        requests_post_mock.side_effect = exception_class
        with open('test_last_runs.csv', 'w') as last_runs:
            last_runs.write(CSV_FILE)

        # write out the lastruns.txt file that would usually be on the archive
        with open('lastrun_wish.txt', 'w') as lastrun_wish:
            lastrun_wish.write(LASTRUN_WISH_TXT)

        # Perform test
        update_last_runs('test_last_runs.csv')
        requests_post_mock.asssert_called_once()

        # Read the CSV and ensure it has been updated
        with open('test_last_runs.csv') as csv_file:
            csv_reader = csv.reader(csv_file)
            for row in csv_reader:
                if row:  # Avoid the empty rows
                    self.assertEqual('44733', row[1])

        assert logger_mock.info.call_count == 3
        assert logger_mock.error.call_count == 2

    @parameterized.expand([
        [ConnectionError],
        [RequestException],
    ])
    @patch('autoreduce_run_detection.run_detection.requests.post', return_value=[44734, 44735])
    @patch('autoreduce_run_detection.run_detection.LOGGING')
    @patch('autoreduce_run_detection.run_detection.TEAMS_URL', return_value="http://fake_url")
    def test_update_last_runs_with_error_and_teams_url_also_fails(self, exception_class, _: Mock, logger_mock: Mock,
                                                                  requests_post_mock: Mock):
        """
        Test trying to update last runs but both the request to the
        autoreduce API and the request to the teams API fail.
        """
        # Setup test
        requests_post_mock.side_effect = exception_class
        with open('test_last_runs.csv', 'w') as last_runs:
            last_runs.write(CSV_FILE)

        # write out the lastruns.txt file that would usually be on the archive
        with open('lastrun_wish.txt', 'w') as lastrun_wish:
            lastrun_wish.write(LASTRUN_WISH_TXT)

        # Perform test
        update_last_runs('test_last_runs.csv')
        requests_post_mock.asssert_called_once()

        # Read the CSV and ensure it has been updated
        with open('test_last_runs.csv') as csv_file:
            csv_reader = csv.reader(csv_file)
            for row in csv_reader:
                if row:  # Avoid the empty rows
                    self.assertEqual('44733', row[1])

        assert logger_mock.info.call_count == 2
        assert logger_mock.error.call_count == 3

    @patch(
        'autoreduce_run_detection.run_detection.requests.post',
        return_value=[44734, 44735],
        side_effect=[RequestException, None]  # this means the second call will NOT raise an exception
    )
    @patch('autoreduce_run_detection.run_detection.LOGGING')
    @patch('autoreduce_run_detection.run_detection.TEAMS_URL', return_value="http://fake_url")
    def test_update_last_runs_with_error_and_teams_url(self, teams_url: Mock, logger_mock: Mock,
                                                       requests_post_mock: Mock):
        """
        Test trying to update last runs but the request to the autoreduce API fails.
        The request to the teams API does not raise an exception in this test, i.e. the success path.
        """
        # Setup test
        with open('test_last_runs.csv', 'w') as last_runs:
            last_runs.write(CSV_FILE)

        # write out the lastruns.txt file that would usually be on the archive
        with open('lastrun_wish.txt', 'w') as lastrun_wish:
            lastrun_wish.write(LASTRUN_WISH_TXT)

        # Perform test
        update_last_runs('test_last_runs.csv')
        requests_post_mock.asssert_called_once()

        # Read the CSV and ensure it has been updated
        with open('test_last_runs.csv') as csv_file:
            csv_reader = csv.reader(csv_file)
            for row in csv_reader:
                if row:  # Avoid the empty rows
                    self.assertEqual('44733', row[1])

        assert logger_mock.info.call_count == 2
        assert logger_mock.error.call_count == 2

        assert requests_post_mock.call_count == 2
        assert teams_url in requests_post_mock.call_args[0]

    @staticmethod
    @patch('autoreduce_run_detection.run_detection.update_last_runs')
    def test_main(update_last_runs_mock):
        main()
        update_last_runs_mock.assert_called_with(LOCAL_CACHE_LOCATION)
        update_last_runs_mock.assert_called_once()

    @staticmethod
    @patch('autoreduce_run_detection.run_detection.update_last_runs')
    def test_main_lock_timeout(_):
        with FileLock('{}.lock'.format(LOCAL_CACHE_LOCATION)):
            main()



# ##################################################################################### #
# ISIS File Polling Repository : https://github.com/ISISSoftwareServices/ISISFilePolling
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# ##################################################################################### #
# pylint: skip-file
"""
Settings for End of run monitor
"""
import os
from autoreduce_utils.settings import AUTOREDUCE_HOME_ROOT

LOCAL_CACHE_LOCATION = os.path.join(AUTOREDUCE_HOME_ROOT, 'last_runs.csv')

# Data Cache
CYCLE_FOLDER = "cycle_22_1"

if "AUTOREDUCTION_PRODUCTION" in os.environ:
    AUTOREDUCE_API_URL = "https://reduce.isis.cclrc.ac.uk/api/runs/{instrument}"
else:
    AUTOREDUCE_API_URL = "http://127.0.0.1:8000/api/runs/{instrument}"

AUTOREDUCE_TOKEN = os.environ.get('AUTOREDUCE_TOKEN')

# set this ENV var to allow error notifications to be sent to the Teams support channel
TEAMS_URL = os.environ.get("TEAMS_URL", None)



# ##################################################################################### #
# ISIS File Polling Repository : https://github.com/ISISSoftwareServices/ISISFilePolling
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# ##################################################################################### #
# pylint:skip-file
"""
Setup for the project
"""

from setuptools import setup, find_packages

setup_requires = [
    'autoreduce_utils>=22.0.0.dev7',
    'filelock',
    'requests',
]

setup(name='autoreduce-run-detection',
      version='22.0.0.dev4',
      description='A service to discover and ingest new files produced by the ISIS neutron and muon facility',
      author='ISIS Autoreduction Team',
      url='https://github.com/ISISScientificComputing/autoreduce-run-detection',
      install_requires=setup_requires,
      packages=find_packages(),
      classifiers=[
          "Programming Language :: Python :: 3 :: Only",
      ],
      entry_points={"console_scripts": ["autoreduce-run-detection = autoreduce_run_detection.run_detection:main"]})
